#ifdef _WIN32
#include <windows.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <GL/gl.h>
#include <GL/glut.h>
#include <AR/gsub.h>
#include <AR/param.h>
#include <AR/ar.h>
#include <AR/video.h>

// EXAMPLES:
//
// char *vconf = NULL;
// char *vconf = "showDlg";
// char *vconf = "deviceName=Microsoft DV Camera,videoWidth=720,videoHeight=576";
// char    *vconf = "ieee1394id=437d3b0201460008,videoWidth=180,videoHeight=144";
// char    *vconf = "showDlg";
   char   *vconf = "showDlg,deviceName=Microsoft DV Camera,deinterlaceState=on,deinterlaceMethod=blend,"
                   "pixelFormat=PIXELFORMAT_RGB32,videoWidth=320,videoHeight=240";

/*  parameter format is either NULL or a list of tokens, separated by commas ","
 *
 *   BINARY TOKENS:
 *   --------------
 *   flipH    : flip image horizontally (WARNING: non-optimal performance)
 *   flipV    : flip image vertically (WARNING: non-optimal performance)
 *   showDlg  : displays either WDM capture filter's property page or
 *              MSDV decoder format dialog (depending on source media type).
 *
 *   PARAMETRIZED TOKENS:
 *   --------------------
 *   videoWidth=?  : preferred video width, EXAMPLE: "videoWidth=720"
 *   videoHeight=? : preferred video height, EXAMPLE: "videoHeight=576"
 *   deviceName=?  : preferred WDM device (WARNING: locale-dependent), 
 *                   i.e. try to match substring <?> with "friendly names" of
 *                   enumerated DirectShow WDM wrappers (ksproxy.ax).
 *                   EXAMPLE: "deviceName=Microsoft DV Camera" for IEEE1394 DV devices
 *                            "deviceName=QuickCam" for Logitech QuickCam
 *   ieee1394id=?  : Unique 64-bit device identifier, as defined by IEEE 1394.
 *					 Hexadecimal value expected, i.e. "ieee1394id=437d3b0201460008"
 *                   Use /bin/IEEE394_id.exe to determine your camera's ID.
 *
 *   EXAMPLES:
 *
 *   arVideoOpen(NULL);
 *   arVideoOpen("showDlg");
 *   arVideoOpen("flipH,flipV,showDlg");
 *   arVideoOpen("videoWidth=640,flipH,videoHeight=480,showDlg");
 *   arVideoOpen("deviceName=Microsoft DV Camera,videoWidth=720,videoHeight=480");
 *   arVideoOpen("deviceName=Logitech,videoWidth=320,videoHeight=240,flipV");
 *   arVideoOpen("deviceName=Microsoft DV Camera,ieee1394id=437d3b0201460008");
 *   ...
 */

/*****************************************************************************/

/* external VRML files */
extern "C" {
//#include "irender.h"
#include "object.h"
#include <AR/vrml97.h>
}

/* Object Data */
char            *model_name = "Data/vrml_data";
ObjectData_T    *object;
int             objectnum;

int             xsize, ysize;
int             thresh = 100;
int             count = 0;

char           *cparam_name    = "Data/camera_para.dat";
ARParam         cparam;
        
static void   init(void);
static void   cleanup(void);
static void   keyEvent( unsigned char key, int x, int y);
static void   mainLoop(void);
int draw( ObjectData_T *object, int objectnum );
static int  draw_object( ObjectData_T *object, double gl_para[16] );
static void	  init_lights(void);

int main()
{
	//initialize applications
    init();

	//start video capture
	arVideoCapStart();

	//start the main event loop
    argMainLoop( NULL, keyEvent, mainLoop );

	return 0;
}

static void   keyEvent( unsigned char key, int x, int y)   
{
    /* quit if the ESC key is pressed */
    if( key == 0x1b ) {
        printf("*** %f (frame/sec)\n", (double)count/arUtilTimer());
        cleanup();
        exit(0);
    }
}

/* main loop */
static void mainLoop(void)
{
    ARUint8         *dataPtr;
    ARMarkerInfo    *marker_info;
    int             marker_num;
    int             i,j, k;

    /* grab a video frame */
	if( (dataPtr = (ARUint8 *)arVideoGetImage()) == NULL ) {
        arUtilSleep(2);
        return;
    }
	
    if( count == 0 ) arUtilTimerReset();  
    count++;

    argDrawMode2D();
    argDispImage( dataPtr, 0,0 );

	/* capture the next video frame */
	arVideoCapNext();

    /* detect the markers in the video frame */
    if( arDetectMarker(dataPtr, thresh, &marker_info, &marker_num) < 0 ) {
        cleanup();
        exit(0);
    }

	/* check for object visibility */
    for( i = 0; i < objectnum; i++ ) {
        k = -1;
        for( j = 0; j < marker_num; j++ ) {
            if( object[i].id == marker_info[j].id ) {
                if( k == -1 ) k = j;
                else {
                    if( marker_info[k].cf < marker_info[j].cf ) k = j;
                }
            }
        }
        if( k == -1 ) {
            object[i].visible = 0;
            continue;
        }

        if( object[i].visible == 0 ) {
            arGetTransMat(&marker_info[k],
                          object[i].marker_center, object[i].marker_width,
                          object[i].trans);
        }
        else {
            arGetTransMatCont(&marker_info[k], object[i].trans,
                          object[i].marker_center, object[i].marker_width,
                          object[i].trans);
        }
        object[i].visible = 1;
    }

	/* draw the virtual objects attached to the cards */
    glClearDepth( 1.0 );
    glClear(GL_DEPTH_BUFFER_BIT);
    glDisable(GL_BLEND);
    draw( object, objectnum );

	/* update the VRML animation */
	arVrml97TimerUpdate();

	/* get ready to grab a new video frame */
	argSwapBuffers();
}

static void init( void )
{
	int i;
	ARParam  wparam;

	/* open the video path */
    if( arVideoOpen( vconf ) < 0 ) exit(0);
    /* find the size of the window */
    if( arVideoInqSize(&xsize, &ysize) < 0 ) exit(0);
    printf("Image size (x,y) = (%d,%d)\n", xsize, ysize);

    /* set the initial camera parameters */
    if( arParamLoad(cparam_name, 1, &wparam) < 0 ) {
        printf("Camera parameter load error !!\n");
        exit(0);
    }
    arParamChangeSize( &wparam, xsize, ysize, &cparam );
    arInitCparam( &cparam );
    printf("*** Camera Parameter ***\n");
    arParamDisp( &cparam );

    /* open the graphics window */
    argInit( &cparam, 2.0, 0, 0, 0, 0 );

	/* load in the object data - trained markers and associated bitmap files */
    if( (object=read_VRMLdata(model_name, &objectnum)) == NULL ) exit(0);
    printf("Objectfile num = %d\n", objectnum);

    /* test render all the VRML objects */
    printf("about to render VRML objects \n");
    glEnable(GL_TEXTURE_2D);

    for ( i = 0; i < objectnum; i++ ) {
      printf("rendering %d \n",i);
      //renderVRMLObject(i);
	  arVrml97Draw( object[i].vrml_id );
    }
    glDisable(GL_TEXTURE_2D);

	/* initialize lights */
    init_lights(); 
}

/* cleanup function called when program exits */
static void cleanup(void)
{
	arVideoCapStop();
    arVideoClose();
    argCleanup();
}

/* draw the the AR objects */
int draw( ObjectData_T *object, int objectnum )
{
    int     i;
    double  gl_para[16];
    
    /* calculate the viewing parameters - gl_para */
    for( i = 0; i < objectnum; i++ ) {
        if( object[i].visible == 0 ) continue;

        argConvGlpara(object[i].trans, gl_para);
        draw_object( &object[i], gl_para );
    }
    
    return(0);
}

/* draw the user object */
static int  draw_object( ObjectData_T *object, double gl_para[16] )
{
    argDrawMode3D();
    argDraw3dCamera( 0, 0 );
    glMatrixMode(GL_MODELVIEW);
    glLoadMatrixd( gl_para );

    if( object->vrml_id >= 0 ) {
        arVrml97Draw( object->vrml_id );
    }
    else {
      printf("unknown object type!!\n");
    }

    argDrawMode2D();

    return 0;
}

/* initialize the lights in the scene */
static void init_lights()
{
    GLfloat light_position[] = { 0, 0, -40.0,  0.0};
    GLfloat light_position1[] = {1.0,1.0,-500.0,0.0};
    GLfloat light_position2[] = {1.0,-500.0,0.0,0.0};
    GLfloat light_position3[] = {-500.0,1.0,0.0,0.0};
 
    //GLfloat ambi[] = {1, 1, 1, 0.1};
	GLfloat ambi[] = {1, 1, 1, 1};
    GLfloat lightZeroColor[]    = {1,1,1, 1};
	
    /* light 0 */
    glLightfv(GL_LIGHT0, GL_POSITION, light_position);
    glLightfv(GL_LIGHT0, GL_AMBIENT, ambi);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, lightZeroColor);  /* light 0 */

    /* light 1 */
    glLightfv(GL_LIGHT1, GL_POSITION, light_position1);
  	glLightfv(GL_LIGHT1, GL_AMBIENT, ambi);
  	glLightfv(GL_LIGHT1, GL_DIFFUSE, lightZeroColor);
    
    /* light 2 */
    
    glLightfv(GL_LIGHT2, GL_POSITION, light_position2);
  	glLightfv(GL_LIGHT2, GL_AMBIENT, ambi);
  	glLightfv(GL_LIGHT2, GL_DIFFUSE, lightZeroColor);
    
    /* light 3 */ 
    glLightfv(GL_LIGHT3, GL_POSITION, light_position3);
	  glLightfv(GL_LIGHT3, GL_AMBIENT, ambi);
  	glLightfv(GL_LIGHT3, GL_DIFFUSE, lightZeroColor);
  
   
    /* light 4 */ 
    glLightfv(GL_LIGHT4, GL_POSITION, light_position3);
    glLightfv(GL_LIGHT4, GL_AMBIENT, ambi);
  	glLightfv(GL_LIGHT4, GL_DIFFUSE, lightZeroColor);
	/* light 5 */
    
    glLightfv(GL_LIGHT5, GL_POSITION, light_position3);
  	glLightfv(GL_LIGHT5, GL_AMBIENT, ambi);
  	glLightfv(GL_LIGHT5, GL_DIFFUSE, lightZeroColor);
    /* light 6 */
    
    glLightfv(GL_LIGHT6, GL_POSITION, light_position3);
  	glLightfv(GL_LIGHT6, GL_AMBIENT, ambi);
  	glLightfv(GL_LIGHT6, GL_DIFFUSE, lightZeroColor);
	/* light 7 */
    
    glLightfv(GL_LIGHT7, GL_POSITION, light_position3);
  	glLightfv(GL_LIGHT7, GL_AMBIENT, ambi);
  	glLightfv(GL_LIGHT7, GL_DIFFUSE, lightZeroColor);
  
    glEnable(GL_LIGHTING);
    
    glEnable(GL_LIGHT2);
    glEnable(GL_LIGHT3);
    glEnable(GL_LIGHT4);
	glEnable(GL_LIGHT5);
	glEnable(GL_LIGHT6);
    //glDepthFunc(GL_GEQUAL);
    //glEnable(GL_DEPTH_TEST);
}
